from statistics import NormalDist
from itertools import accumulate
from functools import cache
from operator import or_, and_

from .shared import *


class AbstractBHV:
    # recall
    # information_entropy(p) = -p*log2(p) - (1 - p)*log2(1 - p)
    # so for some active fractions, it should be a lot cheaper than for others
    # hence a specialized method for 2^-n
    @classmethod
    def rand(cls) -> Self:
        raise NotImplementedError()

    @classmethod
    def nrand(cls, n) -> list[Self]:
        return [cls.rand() for _ in range(n)]

    @classmethod
    def rand2(cls, power=1) -> Self:
        r = cls.rand()
        return r if power == 1 else r & cls.rand2(power - 1)

    @classmethod
    def nrand2(cls, n, power=1) -> list[Self]:
        return [cls.rand2(power) for _ in range(n)]

    @classmethod
    def random(cls, active=0.5) -> Self:
        raise NotImplementedError()

    @classmethod
    def nrandom(cls, n, active=0.5) -> list[Self]:
        return [cls.random(active) for _ in range(n)]

    def select_rand(self, other) -> Self:
        return self.rand().select(self, other)

    def select_rand2(self, other, power_left=1) -> Self:
        return self.rand2(power_left).select(self, other)

    def select_random(self, other, frac_left=0.5) -> Self:
        return self.random(frac_left).select(self, other)

    def randomize_half(self) -> Self:
        return self.select_rand(self.rand())

    def randomize_pow(self, pow_randomize=1) -> Self:
        return self.rand().select_rand2(self, pow_randomize)

    def randomize_frac(self, frac_randomize=0.5) -> Self:
        return self.rand().select_random(self, frac_randomize)

    def flip_half(self) -> Self:
        return self ^ self.rand()

    def flip_pow(self, pow_flip=1) -> Self:
        return self ^ self.rand2(pow_flip)

    def flip_frac(self, frac_flip=0.5) -> Self:
        return self ^ self.random(frac_flip)

    def flip_frac_on(self, flip_on_frac) -> Self:
        return self | self.flip_frac(2*flip_on_frac)

    def flip_frac_off(self, flip_off_frac) -> Self:
        return self & self.flip_frac(2*flip_off_frac)

    def permute(self, permutation_id: int) -> Self:
        raise NotImplementedError()

    @classmethod
    def majority3(cls, x, y, z) -> Self:
        return cls.majority([x, y, z])

    @classmethod
    def majority(cls, vs: list[Self]) -> Self:
        half = len(vs)//2
        disjunction = list(accumulate(vs[:half-1:-1], or_))[1:]
        conjunction = list(accumulate(vs[:half-1:-1], and_))[1:]

        @cache
        def rec(ons, offs):
            if ons + 1 > half:
                return disjunction[-offs - 1]
            if offs + 1 > half:
                return conjunction[-ons - 1]
            return vs[ons + offs].select(
                rec(ons + 1, offs),
                rec(ons, offs + 1)
            )

        return rec(0, 0)

    def __eq__(self, other: Self) -> bool:
        raise NotImplementedError()

    def __xor__(self, other: Self) -> Self:
        raise NotImplementedError()

    def __and__(self, other: Self) -> Self:
        raise NotImplementedError()

    def __or__(self, other: Self) -> Self:
        raise NotImplementedError()

    def __invert__(self) -> Self:
        raise NotImplementedError()

    def select(self, when1: Self, when0: Self) -> Self:
        return when0 ^ (self & (when0 ^ when1))

    def mix(self, other: Self, pick_l=0.5) -> Self:
        return self.random(pick_l).select(self, other)

    def active(self) -> int:
        raise NotImplementedError()

    def bias_rel(self, other: Self, rel: Self) -> float:
        rel_l = rel.select(self, self.ZERO).active()
        rel_r = rel.select(other, self.ZERO).active()
        return rel_l / (rel_l + rel_r)

    def active_fraction(self) -> float:
        return self.active()/DIMENSION

    def hamming(self, other: Self) -> int:
        return (self ^ other).active()

    def bit_error_rate(self, other: Self) -> float:
        return (self ^ other).active_fraction()

    def jaccard(self, other: Self) -> float:
        return 1. - float((self & other).active()) / float((self | other).active() + 1E-7)

    def cosine(self, other: Self) -> float:
        return 1 - float((self & other).active()) / float(self.active() * other.active() + 1E-7)**.5

    def std_apart(self, other: Self) -> float:
        p = 0.5
        n = NormalDist(0, (DIMENSION*p*(1 - p))**.5)
        return n.zscore(self.hamming(other))

    def zscore(self) -> float:
        p = 0.5
        n = NormalDist(DIMENSION*p, (DIMENSION*p*(1 - p))**.5)
        return n.zscore(self.active())

    def pvalue(self) -> float:
        p = 0.5
        n = NormalDist(DIMENSION*p, (DIMENSION*p*(1 - p))**.5)
        s = n.cdf(self.active())
        return 2.*min(s, 1. - s)

    def sixsigma(self, other: Self) -> bool:
        return abs(self.std_apart(other)) < 6

    # Alternative implementations

    @classmethod
    def _majority3_select(cls, a: Self, b: Self, c: Self) -> Self:
        # C:  1 0 0 1 0 1 1

        # |:  1 1 1 1 0 1 0
        # &:  1 0 0 0 0 0 0

        # M:  1 0 0 1 0 1 0

        return a.select(b | c, b & c)

    @classmethod
    def _majority5_select(cls, x4: Self, x3: Self, x2: Self, x1: Self, x0: Self) -> Self:
        t1 = x1 | x0
        b1 = x1 & x0
        m2 = x2.select(t1, b1)
        t2 = x2 | t1
        b2 = x2 & b1
        m3h = x3.select(t2, m2)
        m3l = x3.select(m2, b2)
        m4 = x4.select(m3h, m3l)
        return m4

    @classmethod
    def _majority7_select(cls, x6: Self, x5: Self, x4: Self, x3: Self, x2: Self, x1: Self, x0: Self) -> Self:
        t1 = x1 | x0
        b1 = x1 & x0
        m2 = x2.select(t1, b1)
        t2 = x2 | t1
        b2 = x2 & b1
        m3h = x3.select(t2, m2)
        m3l = x3.select(m2, b2)
        m4 = x4.select(m3h, m3l)
        t3 = x3 | t2
        b3 = x3 & b2
        m4h = x4.select(t3, m3h)
        m4l = x4.select(m3l, b3)
        m5h = x5.select(m4h, m4)
        m5l = x5.select(m4, m4l)
        m6 = x6.select(m5h, m5l)
        return m6

    @classmethod
    def _majority9_select(cls, x8: Self, x7: Self, x6: Self, x5: Self, x4: Self, x3: Self, x2: Self, x1: Self, x0: Self) -> Self:
        t1 = x0 | x1
        b1 = x0 & x1
        m2 = x2.select(t1, b1)
        t2 = x2 | t1
        b2 = x2 & b1
        m3h = x3.select(t2, m2)
        m3l = x3.select(m2, b2)
        t3 = x3 | t2
        b3 = x3 & b2
        m4h = x4.select(t3, m3h)
        m4 = x4.select(m3h, m3l)
        m4l = x4.select(m3l, b3)
        t4 = x4 | t3
        b4 = x4 & b3
        m5hh = x5.select(t4, m4h)
        m5h = x5.select(m4h, m4)
        m5l = x5.select(m4, m4l)
        m5ll = x5.select(m4l, b4)
        m6h = x6.select(m5hh, m5h)
        m6 = x6.select(m5h, m5l)
        m6l = x6.select(m5l, m5ll)
        m7h = x7.select(m6h, m6)
        m7l = x7.select(m6, m6l)
        m8 = x8.select(m7h, m7l)
        return m8

    @classmethod
    def _majority3(cls, a: Self, b: Self, c: Self) -> Self:
        # a:  1 0 0 1 0 1 1
        # b:  1 1 0 1 0 1 0
        # c:  1 0 1 0 0 0 0
        # M:  1 0 0 1 0 1 0

        # at least 2/3 agreeing on TRUE
        abh = a & b
        bch = b & c
        cah = c & a
        h = abh | bch | cah
        return h

    @classmethod
    def _majority5(cls, a: Self, b: Self, c: Self, d: Self, e: Self) -> Self:
        # at least 3/5 agreeing on TRUE

        # 2*10 AND
        # 9*1 OR
        # (b ∧ d ∧ e) ∨ (a ∧ d ∧ e) ∨ (b ∧ c ∧ e) ∨ (a ∧ c ∧ e) ∨ (b ∧ c ∧ d) ∨ (a ∧ c ∧ d) ∨ (c ∧ d ∧ e) ∨ (a ∧ b ∧ e) ∨ (a ∧ b ∧ d) ∨ (a ∧ b ∧ c)

        ab = a & b
        cd = c & d
        de = d & e
        ce = c & e

        bde = b & de
        ade = a & de

        bce = b & ce
        ace = a & ce

        bcd = b & cd
        acd = a & cd
        cde = e & cd

        abe = ab & e
        abd = ab & d
        abc = ab & c

        h = bde | ade | bce | ace | bcd | acd | cde | abe | abd | abc
        return h

    @classmethod
    def _majority5_via_3(cls, a: Self, b: Self, c: Self, d: Self, e: Self) -> Self:
        return cls._majority3(a, cls._majority3(b, c, d), cls._majority3(e, d, cls._majority3(c, b, a)))

    @classmethod
    def _majority7_via_3(cls, a: Self, b: Self, c: Self, d: Self, e: Self, f: Self, g: Self) -> Self:
        mdef = cls._majority3(d, e, f)
        mabc = cls._majority3(a, b, c)
        return cls._majority3(g, cls._majority3(c, mdef, cls._majority3(a, b, mdef)), cls._majority3(f, mabc, cls._majority3(d, e, mabc)))

    ZERO: Self
    ONE: Self


class MemoizedPermutation:
    _permutations: dict[int | tuple[int, ...], Self]

    @classmethod
    def random(cls) -> Self:
        raise NotImplementedError()

    def __mul__(self, other: Self) -> Self:
        raise NotImplementedError()

    def __invert__(self) -> Self:
        raise NotImplementedError()

    def __pow__(self, power):
        permutation = self
        for _ in range(power - 1):
            permutation = permutation*permutation
        return permutation

    def __call__(self, hv: 'AbstractBHV') -> 'AbstractBHV':
        raise NotImplementedError()

    @classmethod
    def _get_singular(cls, permutation_id: int) -> Self:
        if permutation_id in cls._permutations:
            return cls._permutations[permutation_id]
        elif -permutation_id in cls._permutations:
            inv_permutation = cls._permutations[-permutation_id]
            permutation = ~inv_permutation
            cls._permutations[permutation_id] = permutation
            return permutation
        else:
            permutation = cls.random()
            cls._permutations[permutation_id] = permutation
            return permutation

    @classmethod
    def _get_composite(cls, permutation_id: tuple[int, ...]) -> Self:
        # this can be optimized a lot by looking for partial compositions
        composite_permutation = cls._get_singular(permutation_id[0])
        for component_permutation_id in permutation_id[1:]:
            component_permutation = cls.get(component_permutation_id)
            composite_permutation = composite_permutation * component_permutation

        cls._permutations[permutation_id] = composite_permutation
        return composite_permutation

    @classmethod
    def get(cls, permutation_id: int | tuple[int, ...]) -> Self:
        permutation = cls._get_singular(permutation_id) if isinstance(permutation_id, int) else \
            cls._get_composite(permutation_id)
        return permutation
